/*
 *  linux/fs/file_table.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */

#include <linuxmt/types.h>
#include <linuxmt/config.h>
#include <linuxmt/fs.h>
#include <linuxmt/string.h>
#include <linuxmt/mm.h>

/*
 * first_file points to a doubly linked list of all file structures in
 *            the system.
 * nr_files   holds the length of this list.
 */

struct file * first_file = NULL;
int nr_files = 0;
int max_files = NR_FILE;
static struct file file_array[NR_FILE];


/*
 * Insert a new file structure at the head of the list of available ones.
 */
 
static void insert_file_free(file)
struct file *file;
{
	file->f_count = 0;
	file->f_next = first_file;
	if(first_file)
	{
		file->f_prev = first_file->f_prev;
		file->f_prev->f_next = file;
	}
	else
		file->f_prev = NULL;
	file->f_next->f_prev = file;
	first_file = file;
}

/*
 * Remove a file structure from the list of available ones.
 */

static void remove_file_free(file)
struct file *file;
{
	if (first_file == file)
		first_file = first_file->f_next;
	file->f_next->f_prev = file->f_prev;
	file->f_prev->f_next = file->f_next;
	file->f_next = file->f_prev = NULL;
}

/*
 * Insert a file structure at the end of the list of available ones.
 */
 
static void put_last_free(file)
struct file *file;
{
	file->f_prev = first_file->f_prev;
	file->f_prev->f_next = file;
	file->f_next = first_file;
	file->f_next->f_prev = file;
}

void file_table_init()
{
	struct file * file;
	int i=NR_FILE;

	file = &file_array[0];

	nr_files += i;

	for (; i ; i--)
		insert_file_free(file++);
}


/*
 * Find an unused file structure and return a pointer to it.
 * Returns NULL, if there are no more free file structures or
 * we run out of memory.
 */
 
struct file * get_empty_filp()
{
	int i;
	struct file * f;

	if (!first_file)
		return NULL;

	do {
		for (f = first_file, i=0; i < nr_files; i++, f = f->f_next)
			if (!f->f_count) {
				remove_file_free(f);
				memset(f,0,sizeof(*f));
				put_last_free(f);
				f->f_count = 1;
				f->f_version = ++event;
				return f;
			}
	} while (nr_files < max_files);

	return NULL;
}

